use galactica_content::{Content, SystemHandle, SystemObjectHandle};
use rapier2d::geometry::ColliderHandle;

use crate::{camera::Camera, inputstatus::InputStatus, PlayerStatus};

/// What the player has selected
#[derive(Debug, Clone, Copy)]
pub enum PlayerSelection {
	/// We have nothing selected
	None,

	/// We have a system body selected
	OrbitingBody(SystemObjectHandle),

	/// We have a ship selected
	Ship(ColliderHandle),
}

impl PlayerSelection {
	pub fn get_planet(&self) -> Option<SystemObjectHandle> {
		match self {
			Self::OrbitingBody(h) => Some(*h),
			_ => None,
		}
	}
}

#[derive(Debug)]
pub struct PlayerAgent {
	/// Which ship this player is controlling
	pub ship: Option<ColliderHandle>,

	/// What the player has selected
	pub selection: PlayerSelection,

	/// This player's camera
	pub camera: Camera,

	/// What buttons this player is pressing
	pub input: InputStatus,
}

impl PlayerAgent {
	pub fn new(ship: ColliderHandle) -> Self {
		Self {
			input: InputStatus::new(),
			camera: Camera::new(),
			ship: Some(ship),
			selection: PlayerSelection::OrbitingBody(SystemObjectHandle {
				system_handle: SystemHandle { index: 0 },
				body_index: 1,
			}),
		}
	}

	pub fn set_camera_aspect(&mut self, v: f32) {
		self.camera.aspect = v
	}

	pub fn step(&mut self, ct: &Content, status: PlayerStatus) {
		if self.input.get_v_scroll() != 0.0 {
			self.camera.zoom = (self.camera.zoom + self.input.get_v_scroll())
				.clamp(ct.get_config().zoom_min, ct.get_config().zoom_max);
		}

		if status.pos.is_some() {
			self.camera.pos = status.pos.unwrap();
		}
	}
}
