use glyphon::{cosmic_text::Align, FamilyOwned, FontSystem, Style, Weight};
use log::error;
use rhai::{FnNamespace, FuncRegistration, ImmutableString, Module};
use std::{cell::RefCell, rc::Rc};

use super::super::{Color, Rect};
use crate::ui::{elements::TextBox, UiElement, UiState};

pub fn build_textbox_module(
	font_src: Rc<RefCell<FontSystem>>,
	state_src: Rc<RefCell<UiState>>,
) -> Module {
	let mut module = Module::new();
	module.set_id("GalacticaTextboxModule");

	let state = state_src.clone();
	let font = font_src.clone();
	let _ = FuncRegistration::new("add")
		.with_namespace(FnNamespace::Internal)
		.set_into_module(
			&mut module,
			// TODO: fix ugly spaces
			move |name: ImmutableString,
			      font_size: f32,
			      line_height: f32,
			      rect: Rect,
			      color: Color| {
				let mut ui_state = state.borrow_mut();

				if ui_state.elements.contains_key(&name) {
					error!("tried to make a textbox using an existing name `{name}`");
					return;
				}

				ui_state.names.push(name.clone());
				ui_state.elements.insert(
					name.clone(),
					UiElement::Text(TextBox::new(
						&mut font.borrow_mut(),
						name.clone(),
						font_size,
						line_height,
						rect,
						color,
					)),
				);
			},
		);

	let state = state_src.clone();
	let font = font_src.clone();
	let _ = FuncRegistration::new("set_text")
		.with_namespace(FnNamespace::Internal)
		.set_into_module(
			&mut module,
			move |name: ImmutableString, text: ImmutableString| {
				let mut ui_state = state.borrow_mut();
				match ui_state.get_mut_by_name(&name) {
					Some(UiElement::Text(x)) => x.set_text(&mut font.borrow_mut(), text.as_str()),
					_ => {
						error!("called `textbox::set_text` on an invalid name `{name}`")
					}
				}
			},
		);

	let state = state_src.clone();
	let font = font_src.clone();
	let _ = FuncRegistration::new("align_left")
		.with_namespace(FnNamespace::Internal)
		.set_into_module(&mut module, move |name: ImmutableString| {
			let mut ui_state = state.borrow_mut();
			match ui_state.get_mut_by_name(&name) {
				Some(UiElement::Text(x)) => x.set_align(&mut font.borrow_mut(), Align::Left),
				_ => {
					error!("called `textbox::align_left` on an invalid name `{name}`")
				}
			}
		});

	let state = state_src.clone();
	let font = font_src.clone();
	let _ = FuncRegistration::new("align_right")
		.with_namespace(FnNamespace::Internal)
		.set_into_module(&mut module, move |name: ImmutableString| {
			let mut ui_state = state.borrow_mut();
			match ui_state.get_mut_by_name(&name) {
				Some(UiElement::Text(x)) => x.set_align(&mut font.borrow_mut(), Align::Right),
				_ => {
					error!("called `textbox::align_right` on an invalid name `{name}`")
				}
			}
		});

	let state = state_src.clone();
	let font = font_src.clone();
	let _ = FuncRegistration::new("align_justify")
		.with_namespace(FnNamespace::Internal)
		.set_into_module(&mut module, move |name: ImmutableString| {
			let mut ui_state = state.borrow_mut();
			match ui_state.get_mut_by_name(&name) {
				Some(UiElement::Text(x)) => x.set_align(&mut font.borrow_mut(), Align::Justified),
				_ => {
					error!("called `textbox::align_justify` on an invalid name `{name}`")
				}
			}
		});

	let state = state_src.clone();
	let font = font_src.clone();
	let _ = FuncRegistration::new("align_center")
		.with_namespace(FnNamespace::Internal)
		.set_into_module(&mut module, move |name: ImmutableString| {
			let mut ui_state = state.borrow_mut();
			match ui_state.get_mut_by_name(&name) {
				Some(UiElement::Text(x)) => x.set_align(&mut font.borrow_mut(), Align::Center),
				_ => {
					error!("called `textbox::align_center` on an invalid name `{name}`")
				}
			}
		});

	let state = state_src.clone();
	let font = font_src.clone();
	let _ = FuncRegistration::new("weight_bold")
		.with_namespace(FnNamespace::Internal)
		.set_into_module(&mut module, move |name: ImmutableString| {
			let mut ui_state = state.borrow_mut();
			match ui_state.get_mut_by_name(&name) {
				Some(UiElement::Text(x)) => x.set_weight(&mut font.borrow_mut(), Weight::BOLD),
				_ => {
					error!("called `textbox::weight_bold` on an invalid name `{name}`")
				}
			}
		});

	let state = state_src.clone();
	let font = font_src.clone();
	let _ = FuncRegistration::new("weight_normal")
		.with_namespace(FnNamespace::Internal)
		.set_into_module(&mut module, move |name: ImmutableString| {
			let mut ui_state = state.borrow_mut();
			match ui_state.get_mut_by_name(&name) {
				Some(UiElement::Text(x)) => x.set_weight(&mut font.borrow_mut(), Weight::NORMAL),
				_ => {
					error!("called `textbox::weight_normal` on an invalid name `{name}`")
				}
			}
		});

	let state = state_src.clone();
	let font = font_src.clone();
	let _ = FuncRegistration::new("font_serif")
		.with_namespace(FnNamespace::Internal)
		.set_into_module(&mut module, move |name: ImmutableString| {
			let mut ui_state = state.borrow_mut();
			match ui_state.get_mut_by_name(&name) {
				Some(UiElement::Text(x)) => x.set_font(&mut font.borrow_mut(), FamilyOwned::Serif),
				_ => {
					error!("called `textbox::font_serif` on an invalid name `{name}`")
				}
			}
		});

	let state = state_src.clone();
	let font = font_src.clone();
	let _ = FuncRegistration::new("font_sans")
		.with_namespace(FnNamespace::Internal)
		.set_into_module(&mut module, move |name: ImmutableString| {
			let mut ui_state = state.borrow_mut();
			match ui_state.get_mut_by_name(&name) {
				Some(UiElement::Text(x)) => {
					x.set_font(&mut font.borrow_mut(), FamilyOwned::SansSerif)
				}
				_ => {
					error!("called `textbox::font_sans` on an invalid name `{name}`")
				}
			}
		});

	let state = state_src.clone();
	let font = font_src.clone();
	let _ = FuncRegistration::new("font_mono")
		.with_namespace(FnNamespace::Internal)
		.set_into_module(&mut module, move |name: ImmutableString| {
			let mut ui_state = state.borrow_mut();
			match ui_state.get_mut_by_name(&name) {
				Some(UiElement::Text(x)) => {
					x.set_font(&mut font.borrow_mut(), FamilyOwned::Monospace)
				}
				_ => {
					error!("called `textbox::font_mono` on an invalid name `{name}`")
				}
			}
		});

	let state = state_src.clone();
	let font = font_src.clone();
	let _ = FuncRegistration::new("style_normal")
		.with_namespace(FnNamespace::Internal)
		.set_into_module(&mut module, move |name: ImmutableString| {
			let mut ui_state = state.borrow_mut();
			match ui_state.get_mut_by_name(&name) {
				Some(UiElement::Text(x)) => x.set_style(&mut font.borrow_mut(), Style::Normal),
				_ => {
					error!("called `textbox::style_normal` on an invalid name `{name}`")
				}
			}
		});

	let state = state_src.clone();
	let font = font_src.clone();
	let _ = FuncRegistration::new("style_italic")
		.with_namespace(FnNamespace::Internal)
		.set_into_module(&mut module, move |name: ImmutableString| {
			let mut ui_state = state.borrow_mut();
			match ui_state.get_mut_by_name(&name) {
				Some(UiElement::Text(x)) => x.set_style(&mut font.borrow_mut(), Style::Italic),
				_ => {
					error!("called `textbox::style_italic` on an invalid name `{name}`")
				}
			}
		});

	return module;
}
