use galactica_content::{Ship, SystemObject, SystemObjectHandle};
use galactica_system::{
	data::{self},
	phys::{objects::PhysShip, PhysSimShipHandle},
};
use galactica_util::to_degrees;
use log::error;
use rapier2d::dynamics::RigidBody;
use rhai::{Array, CustomType, Dynamic, ImmutableString, TypeBuilder};
use std::sync::Arc;

use super::{Color, UiVector};
use crate::{RenderInput, RenderState};

#[derive(Debug, Clone)]
pub struct ShipState {
	ship: Option<PhysSimShipHandle>,
	input: Arc<RenderInput>,
}

impl ShipState {
	fn get_content(&mut self) -> &Ship {
		let ship = self
			.input
			.phys_img
			.get_ship(self.ship.as_ref().unwrap())
			.unwrap();
		let handle = ship.ship.get_data().get_content();
		self.input.ct.get_ship(handle)
	}

	fn get_ship(&mut self) -> &PhysShip {
		let ship = self
			.input
			.phys_img
			.get_ship(self.ship.as_ref().unwrap())
			.unwrap();
		&ship.ship
	}

	fn get_body(&mut self) -> &RigidBody {
		let ship = self
			.input
			.phys_img
			.get_ship(self.ship.as_ref().unwrap())
			.unwrap();
		&ship.rigidbody
	}

	fn landed_on(&mut self) -> SystemObjectState {
		let input = self.input.clone();
		match self.get_ship().get_data().get_state() {
			data::ShipState::Landed { target } => {
				return SystemObjectState {
					input,
					object: Some(*target),
				}
			}
			_ => {
				return SystemObjectState {
					input,
					object: None,
				}
			}
		};
	}
}

impl CustomType for ShipState {
	fn build(mut builder: TypeBuilder<Self>) {
		builder
			.with_name("ShipState")
			.with_fn("is_some", |s: &mut Self| s.ship.is_some())
			.with_fn("is_dead", |s: &mut Self| {
				s.get_ship().get_data().get_state().is_dead()
			})
			.with_fn("is_landed", |s: &mut Self| {
				s.get_ship().get_data().get_state().is_landed()
			})
			.with_fn("is_landing", |s: &mut Self| {
				s.get_ship().get_data().get_state().is_landing()
			})
			.with_fn("is_flying", |s: &mut Self| {
				s.get_ship().get_data().get_state().is_flying()
			})
			.with_fn("is_unlanding", |s: &mut Self| {
				s.get_ship().get_data().get_state().is_unlanding()
			})
			.with_fn("is_collapsing", |s: &mut Self| {
				s.get_ship().get_data().get_state().is_collapsing()
			})
			.with_fn("name", |s: &mut Self| s.get_content().name.clone())
			.with_fn("thumbnail", |s: &mut Self| s.get_content().thumb)
			.with_fn("landed_on", |s: &mut Self| s.landed_on())
			.with_fn("get_shields", |s: &mut Self| {
				s.get_ship().get_data().get_shields()
			})
			.with_fn("get_total_shields", |s: &mut Self| {
				s.get_ship().get_data().get_outfits().get_total_shields()
			})
			.with_fn("get_total_hull", |s: &mut Self| s.get_content().hull)
			.with_fn("get_hull", |s: &mut Self| {
				s.get_ship().get_data().get_hull()
			})
			.with_fn("get_size", |s: &mut Self| s.get_content().size)
			.with_fn("get_uid", |s: &mut Self| format!("{}", s.get_ship().uid))
			.with_fn("get_pos", |s: &mut Self| {
				let t = s.get_body().translation();
				UiVector::new(t.x, t.y)
			})
			.with_fn("get_faction_color", |s: &mut Self| {
				let h = s.get_ship().get_data().get_faction();
				let c = s.input.ct.get_faction(h).color;
				Color::new(c[0], c[1], c[2], 1.0)
			});
	}
}

#[derive(Debug, Clone)]
pub struct SystemObjectState {
	object: Option<SystemObjectHandle>,
	input: Arc<RenderInput>,
}

impl SystemObjectState {
	fn get_content(&mut self) -> &SystemObject {
		self.input.ct.get_system_object(self.object.unwrap())
	}
}

impl CustomType for SystemObjectState {
	fn build(mut builder: TypeBuilder<Self>) {
		builder
			.with_name("SystemObjectState")
			//
			// Get landable name
			.with_fn("name", |s: &mut Self| {
				s.get_content()
					.name
					.as_ref()
					.map(|x| x.to_string())
					.unwrap_or_else(|| {
						error!("UI called `name()` on a system object which doesn't provide one");
						"".to_string()
					})
			})
			//
			// Get landable description
			.with_fn("desc", |s: &mut Self| {
				s.get_content()
					.desc
					.as_ref()
					.map(|x| x.to_string())
					.unwrap_or_else(|| {
						error!("UI called `name()` on a system object which doesn't provide one");
						"".to_string()
					})
			})
			//
			// Get landable landscape image
			.with_fn("image", |s: &mut Self| {
				let handle = s.get_content().image;
				if let Some(handle) = handle {
					s.input.ct.get_sprite(handle).name.clone()
				} else {
					error!("UI called `image()` on a system object which doesn't provide one");
					"".to_string()
				}
			})
			.with_fn("is_some", |s: &mut Self| s.object.is_some())
			.with_fn("==", |a: &mut Self, b: Self| a.object == b.object)
			.with_fn("get_size", |s: &mut Self| s.get_content().size)
			.with_fn("get_label", |s: &mut Self| {
				ImmutableString::from(&s.get_content().label)
			})
			.with_fn("get_angle", |s: &mut Self| {
				to_degrees(s.get_content().angle)
			})
			.with_fn("get_pos", |s: &mut Self| {
				let t = s.get_content().pos;
				UiVector::new(t.x, t.y)
			})
			.with_fn("get_pos_z", |s: &mut Self| {
				let t = s.get_content().pos;
				t.z
			});
	}
}

#[derive(Debug, Clone)]
pub struct State {
	input: Arc<RenderInput>,
	window_aspect: f32,
}

// TODO: remove this
unsafe impl Send for State {}
unsafe impl Sync for State {}

impl State {
	pub fn new(state: &RenderState, input: Arc<RenderInput>) -> Self {
		Self {
			input,
			window_aspect: state.window_aspect,
		}
	}

	pub fn player_ship(&mut self) -> ShipState {
		ShipState {
			input: self.input.clone(),
			ship: self.input.player.ship.map(|x| PhysSimShipHandle(x)),
		}
	}

	pub fn ships(&mut self) -> Array {
		let mut a = Array::new();
		for s in self.input.phys_img.iter_ships() {
			a.push(Dynamic::from(ShipState {
				input: self.input.clone(),
				ship: Some(PhysSimShipHandle(s.ship.collider)),
			}));
		}
		return a;
	}

	pub fn objects(&mut self) -> Array {
		let mut a = Array::new();
		let s = self.input.current_system;
		for o in &self.input.ct.get_system(s).objects {
			a.push(Dynamic::from(SystemObjectState {
				input: self.input.clone(),
				object: Some(o.handle),
			}));
		}
		return a;
	}
}

impl CustomType for State {
	fn build(mut builder: TypeBuilder<Self>) {
		builder
			.with_name("State")
			.with_fn("player_ship", Self::player_ship)
			.with_fn("ships", Self::ships)
			.with_fn("objects", Self::objects)
			.with_fn("window_aspect", |s: &mut Self| s.window_aspect)
			.with_fn("camera_zoom", |s: &mut Self| s.input.camera_zoom);
	}
}
