use anyhow::{Context, Result};
use galactica_content::Content;
use galactica_system::phys::PhysSimShipHandle;
use galactica_util::rhai_error_to_anyhow;
use log::debug;
use rhai::{
	packages::{BasicArrayPackage, BasicStringPackage, LogicPackage, MoreStringPackage, Package},
	Dynamic, Engine, ImmutableString, Scope,
};
use std::{cell::RefCell, num::NonZeroU32, rc::Rc, sync::Arc};

use super::{
	api::{self, MouseClickEvent, MouseHoverEvent, PlayerShipStateEvent},
	event::Event,
	UiConfig, UiElement, UiState,
};
use crate::{ui::api::State, RenderInput, RenderState};

pub(crate) struct UiScriptExecutor {
	engine: Engine,
	scope: Scope<'static>,

	pub state: Rc<RefCell<UiState>>,

	last_player_state: u32,
	last_scene: Option<ImmutableString>,
}

impl UiScriptExecutor {
	pub fn new(ct: Arc<Content>, state: &mut RenderState) -> Self {
		let scope = Scope::new();
		let elements = Rc::new(RefCell::new(UiState::new(ct.clone(), state)));

		let mut engine = Engine::new_raw();

		// Required for array iteration
		// We may need to add more packages here later.
		engine.register_global_module(BasicArrayPackage::new().as_shared_module());
		engine.register_global_module(LogicPackage::new().as_shared_module());
		engine.register_global_module(BasicStringPackage::new().as_shared_module());
		engine.register_global_module(MoreStringPackage::new().as_shared_module());

		engine.set_max_expr_depths(0, 0);
		// Enables custom operators
		engine.set_fast_operators(false);

		api::register_into_engine(
			&mut engine,
			ct.clone(),
			state.text_font_system.clone(),
			elements.clone(),
		);

		Self {
			engine,
			scope,
			state: elements,
			last_scene: None,
			last_player_state: 0,
		}
	}

	pub fn get_config(&self) -> UiConfig {
		(*self.state).borrow().config.clone()
	}

	/// Change the current scene
	pub fn set_scene(&mut self, state: &RenderState, input: Arc<RenderInput>) -> Result<()> {
		let current_scene = (*self.state).borrow().get_scene().clone();
		if self.last_scene == current_scene {
			return Ok(());
		}
		self.last_scene = current_scene.clone();

		if current_scene.is_none() {
			return Ok(());
		}

		debug!(
			"switched to {}, running `init()`",
			current_scene.as_ref().unwrap()
		);

		self.scope.clear();

		// Drop this right away, since all script calls borrow elm mutably.
		let mut elm = self.state.borrow_mut();
		elm.clear();
		drop(elm);
		let ct = (*self.state).borrow().ct.clone();

		rhai_error_to_anyhow(
			self.engine.call_fn(
				&mut self.scope,
				ct.get_config()
					.ui_scenes
					.get(current_scene.as_ref().unwrap().as_str())
					.unwrap(),
				"init",
				(State::new(state, input.clone()),),
			),
		)
		.with_context(|| format!("while running `init()`"))
		.with_context(|| format!("in ui scene `{}`", current_scene.as_ref().unwrap()))?;

		return Ok(());
	}

	/// Draw all ui elements
	pub fn draw(&mut self, state: &mut RenderState, input: Arc<RenderInput>) -> Result<()> {
		let ct = (*self.state).borrow().ct.clone();

		// Initialize start scene if we haven't yet
		if (*self.state).borrow().get_scene().is_none() {
			(*self.state)
				.borrow_mut()
				.set_scene(ImmutableString::from(&ct.get_config().start_ui_scene));
		}
		self.set_scene(state, input.clone())?;
		let current_scene = (*self.state).borrow().get_scene().clone();

		(*self.state).borrow_mut().step(state, input.clone());

		// Run step() (if it is defined)

		let ast = ct
			.get_config()
			.ui_scenes
			.get(current_scene.as_ref().unwrap().as_str())
			.unwrap();
		if ast.iter_functions().any(|x| x.name == "step") {
			rhai_error_to_anyhow(self.engine.call_fn(
				&mut self.scope,
				ast,
				"step",
				(State::new(state, input.clone()),),
			))
			.with_context(|| format!("while calling `step()`"))
			.with_context(|| format!("in ui scene `{}`", current_scene.as_ref().unwrap()))?;
		}

		// Send player state change events
		if {
			let player = input.player.ship;
			if let Some(player) = player {
				let ship = input.phys_img.get_ship(&PhysSimShipHandle(player)).unwrap();
				if self.last_player_state == 0
					|| NonZeroU32::new(self.last_player_state).unwrap()
						!= ship.ship.get_data().get_state().as_int()
				{
					self.last_player_state = ship.ship.get_data().get_state().as_int().into();
					true
				} else {
					false
				}
			} else {
				self.last_player_state = 0;
				true
			}
		} {
			rhai_error_to_anyhow(
				self.engine.call_fn(
					&mut self.scope,
					ct.get_config()
						.ui_scenes
						.get(current_scene.as_ref().unwrap().as_str())
						.unwrap(),
					"event",
					(State::new(state, input.clone()), PlayerShipStateEvent {}),
				),
			)
			.with_context(|| format!("while handling player state change event"))
			.with_context(|| format!("in ui scene `{}`", current_scene.as_ref().unwrap()))?;
		}

		let len = (*self.state).borrow().len();
		for i in 0..len {
			let event_arg = match (*self.state).borrow_mut().get_mut_by_idx(i).unwrap() {
				UiElement::Sprite(sprite) => {
					// Draw and update sprites
					sprite.step(&input, state);
					sprite.push_to_buffer(&input, state);
					let event = sprite.check_events(&input, state);

					match event {
						Event::None => None,

						Event::MouseClick => Some(Dynamic::from(MouseClickEvent {
							down: true,
							element: sprite.name.clone(),
						})),

						Event::MouseRelease => Some(Dynamic::from(MouseClickEvent {
							down: false,
							element: sprite.name.clone(),
						})),

						Event::MouseHover => Some(Dynamic::from(MouseHoverEvent {
							enter: true,
							element: sprite.name.clone(),
						})),

						Event::MouseUnhover => Some(Dynamic::from(MouseHoverEvent {
							enter: false,
							element: sprite.name.clone(),
						})),
					}
				}

				UiElement::RadialBar(x) => {
					// Draw and update radialbar
					x.step(&input, state);
					x.push_to_buffer(&input, state);
					None
				}

				UiElement::Text(..) => None,
			};

			if let Some(event_arg) = event_arg {
				rhai_error_to_anyhow(
					self.engine.call_fn(
						&mut self.scope,
						ct.get_config()
							.ui_scenes
							.get(current_scene.as_ref().unwrap().as_str())
							.unwrap(),
						"event",
						(State::new(state, input.clone()), event_arg.clone()),
					),
				)
				.with_context(|| format!("while handling event `{:?}`", event_arg))
				.with_context(|| format!("in ui scene `{}`", current_scene.as_ref().unwrap()))?;
			}
		}

		self.scope.rewind(0);
		return Ok(());
	}
}
